@@ -22,6 +22,8 @@
 
 (* Public exceptions and their registration with the C runtime *)
 
+open Pcre_compat
+
 type error =
   | Partial
   | BadPartial
@@ -269,24 +271,30 @@ let regexp_or
   in
   regexp ?study ?limit ?limit_recursion ~iflags ?flags ?chtables big_pat
 
-let string_unsafe_sub s ofs len =
-  let r = String.create len in
-  String.unsafe_blit s ofs r 0 len;
-  r
+let bytes_unsafe_blit_string str str_ofs bts bts_ofs len =
+  let str_bts = Bytes.unsafe_of_string str in
+  Bytes.unsafe_blit str_bts str_ofs bts bts_ofs len
+
+let string_unsafe_sub str ofs len =
+  let res = Bytes.create len in
+  bytes_unsafe_blit_string str ofs res 0 len;
+  Bytes.unsafe_to_string res
 
 let quote s =
   let len = String.length s in
-  let buf = String.create (len lsl 1) in
+  let buf = Bytes.create (len lsl 1) in
   let pos = ref 0 in
   for i = 0 to len - 1 do
     match String.unsafe_get s i with
     | '\\' | '^' | '$' | '.' | '[' | '|'
     | '('  | ')' | '?' | '*' | '+' | '{' as c ->
-      String.unsafe_set buf !pos '\\'; incr pos;
-      String.unsafe_set buf !pos c; incr pos
-    | c -> String.unsafe_set buf !pos c; incr pos
+      Bytes.unsafe_set buf !pos '\\';
+      incr pos;
+      Bytes.unsafe_set buf !pos c;
+      incr pos
+    | c -> Bytes.unsafe_set buf !pos c; incr pos
   done;
-  string_unsafe_sub buf 0 !pos
+  string_unsafe_sub (Bytes.unsafe_to_string buf) 0 !pos
 
 
 (* Matching of patterns and subpattern extraction *)
@@ -371,9 +379,15 @@ let get_named_substring_ofs rex name substrings =
   get_substring_ofs substrings (get_stringnumber rex name)
 
 external unsafe_pcre_exec :
-  irflag -> regexp -> int -> string ->
-  int -> int array -> callout option
-  -> unit = "pcre_exec_stub_bc" "pcre_exec_stub"
+  irflag ->
+  regexp ->
+  pos : int ->
+  subj_start : int ->
+  subj : string ->
+  subgroups2 : int ->
+  int array ->
+  callout option ->
+  unit = "pcre_exec_stub_bc" "pcre_exec_stub"
 
 let make_ovector rex =
   let subgroups1 = capturecount rex + 1 in
@@ -385,7 +399,8 @@ let pcre_exec ?(iflags = 0) ?flags ?(rex = def_rex) ?pat ?(pos = 0)
   let rex = match pat with Some str -> regexp str | _ -> rex in
   let iflags = match flags with Some flags -> rflags flags | _ -> iflags in
   let subgroups2, ovector = make_ovector rex in
-  unsafe_pcre_exec iflags rex pos subj subgroups2 ovector callout;
+  unsafe_pcre_exec
+    iflags rex ~pos ~subj_start:0 ~subj ~subgroups2 ovector callout;
   ovector
 
 let exec ?iflags ?flags ?rex ?pat ?pos ?callout subj =
@@ -564,23 +579,25 @@ let replace ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
     if
       cur_pos > subj_len ||
       try
-        unsafe_pcre_exec iflags rex cur_pos subj subgroups2 ovector callout;
+        unsafe_pcre_exec
+          iflags rex ~pos:cur_pos ~subj_start:0 ~subj
+          ~subgroups2 ovector callout;
         false
       with Not_found -> true
     then
       let postfix_len = max (subj_len - cur_pos) 0 in
       let left = pos + full_len in
-      let res = String.create (left + postfix_len) in
-      String.unsafe_blit subj 0 res 0 pos;
-      String.unsafe_blit subj cur_pos res left postfix_len;
+      let res = Bytes.create (left + postfix_len) in
+      bytes_unsafe_blit_string subj 0 res 0 pos;
+      bytes_unsafe_blit_string subj cur_pos res left postfix_len;
       let inner_coll ofs (templ, ix, len) =
-        String.unsafe_blit templ ix res ofs len; ofs + len in
+        bytes_unsafe_blit_string templ ix res ofs len; ofs + len in
       let coll ofs (res_len, trans_lst) =
         let new_ofs = ofs - res_len in
         let _ = List.fold_left inner_coll new_ofs trans_lst in
         new_ofs in
       let _ = List.fold_left coll left trans_lsts in
-      res
+      Bytes.unsafe_to_string res
     else
       let first = Array.unsafe_get ovector 0 in
       let len = first - cur_pos in
@@ -613,26 +630,28 @@ let qreplace ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
     if
       cur_pos > subj_len ||
       try
-        unsafe_pcre_exec iflags rex cur_pos subj subgroups2 ovector callout;
+        unsafe_pcre_exec
+          iflags rex ~pos:cur_pos ~subj_start:0 ~subj
+          ~subgroups2 ovector callout;
         false
       with Not_found -> true
     then
       let postfix_len = max (subj_len - cur_pos) 0 in
       let left = pos + full_len in
-      let res = String.create (left + postfix_len) in
-      String.unsafe_blit subj 0 res 0 pos;
-      String.unsafe_blit subj cur_pos res left postfix_len;
+      let res = Bytes.create (left + postfix_len) in
+      bytes_unsafe_blit_string subj 0 res 0 pos;
+      bytes_unsafe_blit_string subj cur_pos res left postfix_len;
       let coll ofs = function
         | Some (substr, ix, len) ->
             let new_ofs = ofs - len in
-            String.unsafe_blit substr ix res new_ofs len;
+            bytes_unsafe_blit_string substr ix res new_ofs len;
             new_ofs
         | None ->
             let new_ofs = ofs - templ_len in
-            String.unsafe_blit templ 0 res new_ofs templ_len;
+            bytes_unsafe_blit_string templ 0 res new_ofs templ_len;
             new_ofs in
       let _ = List.fold_left coll left subst_lst in
-      res
+      Bytes.unsafe_to_string res
     else
       let first = Array.unsafe_get ovector 0 in
       let len = first - cur_pos in
@@ -660,21 +679,23 @@ let substitute_substrings ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
     if
       cur_pos > subj_len ||
       try
-        unsafe_pcre_exec iflags rex cur_pos subj subgroups2 ovector callout;
+        unsafe_pcre_exec
+          iflags rex ~pos:cur_pos ~subj_start:0 ~subj
+          ~subgroups2 ovector callout;
         false
       with Not_found -> true
     then
       let postfix_len = max (subj_len - cur_pos) 0 in
       let left = pos + full_len in
-      let res = String.create (left + postfix_len) in
-      String.unsafe_blit subj 0 res 0 pos;
-      String.unsafe_blit subj cur_pos res left postfix_len;
+      let res = Bytes.create (left + postfix_len) in
+      bytes_unsafe_blit_string subj 0 res 0 pos;
+      bytes_unsafe_blit_string subj cur_pos res left postfix_len;
       let coll ofs (templ, ix, len) =
         let new_ofs = ofs - len in
-        String.unsafe_blit templ ix res new_ofs len;
+        bytes_unsafe_blit_string templ ix res new_ofs len;
         new_ofs in
       let _ = List.fold_left coll left subst_lst in
-      res
+      Bytes.unsafe_to_string res
     else
       let first = Array.unsafe_get ovector 0 in
       let len = first - cur_pos in
@@ -715,20 +736,22 @@ let replace_first ?(iflags = 0) ?flags ?(rex = def_rex) ?pat ?(pos = 0)
     failwith "Pcre.replace_first: backreference denotes nonexistent subpattern";
   if with_lp && nsubs = 0 then failwith "Pcre.replace_first: no backreferences";
   try
-    unsafe_pcre_exec iflags rex pos subj subgroups2 ovector callout;
+    unsafe_pcre_exec
+      iflags rex ~pos ~subj_start:0 ~subj
+      ~subgroups2 ovector callout;
     let res_len, trans_lst =
       calc_trans_lst subgroups2 ovector subj templ subst_lst in
     let first = Array.unsafe_get ovector 0 in
     let last = Array.unsafe_get ovector 1 in
     let rest = String.length subj - last in
-    let res = String.create (first + res_len + rest) in
-    String.unsafe_blit subj 0 res 0 first;
+    let res = Bytes.create (first + res_len + rest) in
+    bytes_unsafe_blit_string subj 0 res 0 first;
     let coll ofs (templ, ix, len) =
-      String.unsafe_blit templ ix res ofs len; ofs + len in
+      bytes_unsafe_blit_string templ ix res ofs len; ofs + len in
     let ofs = List.fold_left coll first trans_lst in
-    String.unsafe_blit subj last res ofs rest;
-    res
-  with Not_found -> String.copy subj
+    bytes_unsafe_blit_string subj last res ofs rest;
+    Bytes.unsafe_to_string res
+  with Not_found -> string_copy subj
 
 let qreplace_first ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
                    ?(pos = 0) ?(templ = "") ?callout subj =
@@ -736,18 +759,20 @@ let qreplace_first ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
   let iflags = match flags with Some flags -> rflags flags | _ -> iflags in
   let subgroups2, ovector = make_ovector rex in
   try
-    unsafe_pcre_exec iflags rex pos subj subgroups2 ovector callout;
+    unsafe_pcre_exec
+      iflags rex ~pos ~subj_start:0 ~subj
+      ~subgroups2 ovector callout;
     let first = Array.unsafe_get ovector 0 in
     let last = Array.unsafe_get ovector 1 in
     let len = String.length templ in
     let rest = String.length subj - last in
     let postfix_start = first + len in
-    let res = String.create (postfix_start + rest) in
-    String.unsafe_blit subj 0 res 0 first;
-    String.unsafe_blit templ 0 res first len;
-    String.unsafe_blit subj last res postfix_start rest;
-    res
-  with Not_found -> String.copy subj
+    let res = Bytes.create (postfix_start + rest) in
+    bytes_unsafe_blit_string subj 0 res 0 first;
+    bytes_unsafe_blit_string templ 0 res first len;
+    bytes_unsafe_blit_string subj last res postfix_start rest;
+    Bytes.unsafe_to_string res
+  with Not_found -> string_copy subj
 
 let substitute_substrings_first ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
                                 ?(pos = 0) ?callout ~subst subj =
@@ -755,7 +780,9 @@ let substitute_substrings_first ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
   let iflags = match flags with Some flags -> rflags flags | _ -> iflags in
   let subgroups2, ovector = make_ovector rex in
   try
-    unsafe_pcre_exec iflags rex pos subj subgroups2 ovector callout;
+    unsafe_pcre_exec
+      iflags rex ~pos ~subj_start:0 ~subj
+      ~subgroups2 ovector callout;
     let subj_len = String.length subj in
     let prefix_len = Array.unsafe_get ovector 0 in
     let last = Array.unsafe_get ovector 1 in
@@ -763,12 +790,12 @@ let substitute_substrings_first ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
     let postfix_len = subj_len - last in
     let templ_len = String.length templ in
     let postfix_start = prefix_len + templ_len in
-    let res = String.create (postfix_start + postfix_len) in
-    String.unsafe_blit subj 0 res 0 prefix_len;
-    String.unsafe_blit templ 0 res prefix_len templ_len;
-    String.unsafe_blit subj last res postfix_start postfix_len;
-    res
-  with Not_found -> String.copy subj
+    let res = Bytes.create (postfix_start + postfix_len) in
+    bytes_unsafe_blit_string subj 0 res 0 prefix_len;
+    bytes_unsafe_blit_string templ 0 res prefix_len templ_len;
+    bytes_unsafe_blit_string subj last res postfix_start postfix_len;
+    Bytes.unsafe_to_string res
+  with Not_found -> string_copy subj
 
 let substitute_first ?iflags ?flags ?rex ?pat ?pos
                      ?callout ~subst:str_subst subj =
@@ -785,7 +812,7 @@ let substitute_first ?iflags ?flags ?rex ?pat ?pos
 let internal_psplit flags rex max pos callout subj =
   let subj_len = String.length subj in
   if subj_len = 0 then []
-  else if max = 1 then [String.copy subj]
+  else if max = 1 then [string_copy subj]
   else
     let subgroups2, ovector = make_ovector rex in
 
@@ -812,10 +839,12 @@ let internal_psplit flags rex max pos callout subj =
         (* Checks termination due to max restriction *)
         if cnt = 0 then
           if prematch &&
-             try
-               unsafe_pcre_exec flags rex pos subj subgroups2 ovector callout;
-               true
-             with Not_found -> false
+            try
+              unsafe_pcre_exec
+                flags rex ~pos ~subj_start:pos ~subj
+                ~subgroups2 ovector callout;
+              true
+            with Not_found -> false
           then
             let last = Array.unsafe_get ovector 1 in
             let strs = handle_subgroups strs in
@@ -826,7 +855,9 @@ let internal_psplit flags rex max pos callout subj =
         else
           if
             try
-              unsafe_pcre_exec flags rex pos subj subgroups2 ovector callout;
+              unsafe_pcre_exec
+                flags rex ~pos ~subj_start:pos ~subj
+                ~subgroups2 ovector callout;
               false
             with Not_found -> true
           then string_unsafe_sub subj pos len :: strs
@@ -840,8 +871,8 @@ let internal_psplit flags rex max pos callout subj =
                 else if
                   try
                     unsafe_pcre_exec
-                      (flags lor 0x0410) rex pos subj
-                      subgroups2 ovector callout;
+                      (flags lor 0x0410) rex ~pos ~subj_start:pos ~subj
+                      ~subgroups2 ovector callout;
                     true
                   with Not_found -> false
                 then
@@ -891,13 +922,17 @@ type split_result = Text of string
                   | Group of int * string
                   | NoGroup
 
+let rec strip_all_empty_full = function
+  | Delim _ :: rest -> strip_all_empty_full rest
+  | l -> l
+
 let full_split ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
                ?(pos = 0) ?(max = 0) ?callout subj =
   let rex = match pat with Some str -> regexp str | _ -> rex in
   let iflags = match flags with Some flags -> rflags flags | _ -> iflags in
   let subj_len = String.length subj in
   if subj_len = 0 then []
-  else if max = 1 then [Text (String.copy subj)]
+  else if max = 1 then [Text (string_copy subj)]
   else
     let subgroups2, ovector = make_ovector rex in
 
@@ -927,10 +962,12 @@ let full_split ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
         (* Checks termination due to max restriction *)
         if cnt = 0 then
           if prematch &&
-             try
-               unsafe_pcre_exec iflags rex pos subj subgroups2 ovector callout;
+            try
+              unsafe_pcre_exec
+                iflags rex ~pos ~subj_start:pos ~subj
+                ~subgroups2 ovector callout;
                true
-             with Not_found -> false
+            with Not_found -> false
           then
             let first = Array.unsafe_get ovector 0 in
             let last = Array.unsafe_get ovector 1 in
@@ -945,7 +982,9 @@ let full_split ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
         else
           if
             try
-              unsafe_pcre_exec iflags rex pos subj subgroups2 ovector callout;
+              unsafe_pcre_exec
+                iflags rex ~pos ~subj_start:pos ~subj
+                ~subgroups2 ovector callout;
               false
             with Not_found -> true
           then
@@ -962,8 +1001,8 @@ let full_split ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
                   if
                     try
                       unsafe_pcre_exec
-                        (iflags lor 0x0410) rex pos subj
-                        subgroups2 ovector callout;
+                        (iflags lor 0x0410) rex ~pos ~subj_start:pos ~subj
+                        ~subgroups2 ovector callout;
                       true
                     with Not_found -> false
                   then
@@ -991,7 +1030,8 @@ let full_split ?(iflags = 0) ?flags ?(rex = def_rex) ?pat
                 Text (string_unsafe_sub subj pos (first - pos)) :: strs in
               loop
                 (handle_subgroups (delim :: pre_strs)) (cnt - 1) last false in
-    List.rev (loop [] (max - 1) pos true)
+    let res = loop [] (max - 1) pos true in
+    List.rev (if max = 0 then strip_all_empty_full res else res)
 
 
 (* Additional convenience functions useful in combination with this library *)
